/*
   CEigensystem.F
   diagonalization of a complex n-by-n matrix using the Jacobi algorithm
   code adapted from the "Handbook" routines for complex A
   (Wilkinson, Reinsch: Handbook for Automatic Computation, p. 202)
   this file is part of the Diag library
   last modified 19 Mar 10 th
*/

#include "diag-c.h"


/*
   CEigensystem diagonalizes a general complex n-by-n matrix.
   Input: n, A = n-by-n matrix
   Output: d = vector of eigenvalues, U = transformation matrix
   these fulfill diag(d) = U A U^-1.
*/

void CEigensystem(cInteger n, Complex *A, cInteger ldA,
  Complex *d, Complex *U, cInteger ldU, cInteger sort)
{
  Integer p, q;
  cReal red = .01/(n*n*n*n);
  Complex ev[n][2];

  for( p = 0; p < n; ++p ) {
    ev[p][0] = 0;
    ev[p][1] = d[p] = Element(A,p,p);
  }

  for( p = 0; p < n; ++p ) {
    memset(&Element(U,p,0), 0, n*sizeof(Complex));
    Element(U,p,p) = 1;
  }

  for( nsweeps = 1; nsweeps <= 50; ++nsweeps ) {
    Real thresh = 0;
    for( q = 1; q < n; ++q )
      for( p = 0; p < q; ++p )
        thresh += csq(Element(A,p,q)) + csq(Element(A,q,p));
    if( !(thresh > EPS) ) goto done;

    thresh = (nsweeps < 4) ? thresh*red : 0;

    for( q = 1; q < n; ++q )
      for( p = 0; p < q; ++p ) {
        cComplex Apq = Element(A,p,q);
        cComplex Aqp = Element(A,q,p);
        cReal off = csq(Apq) + csq(Aqp);
        cReal sqp = csq(ev[p][1]);
        cReal sqq = csq(ev[q][1]);
        if( nsweeps > 4 && off < EPS*max(sqp, sqq) ) {
          Element(A,p,q) = 0;
          Element(A,q,p) = 0;
        }
        else if( off > thresh ) {
          Complex delta, x, y, t, s, invc, sx, sy, tx, ty;
          Integer j;

          delta = Apq*Aqp;
          x = .5*(ev[p][1] - ev[q][1]);
          y = csqrt(x*x + delta);
          if( csq(t = x - y) < csq(s = x + y) ) t = s;

          t = 1/t;
          delta *= t;
          ev[p][1] = d[p] + (ev[p][0] += delta);
          ev[q][1] = d[q] + (ev[q][0] -= delta);

          invc = csqrt(delta*t + 1);
          s = t/invc;
          t /= invc + 1;
          sx = s*Apq;
          ty = t*Apq;
          sy = s*Aqp;
          tx = t*Aqp;

          for( j = 0; j < n; ++j ) {
            Complex x = Element(A,j,p);
            Complex y = Element(A,j,q);
            Element(A,j,p) = x + sy*(y - ty*x);
            Element(A,j,q) = y - sx*(x + tx*y);
            x = Element(A,p,j);
            y = Element(A,q,j);
            Element(A,p,j) = x + sx*(y - tx*x);
            Element(A,q,j) = y - sy*(x + ty*y);
          }

          Element(A,p,q) = 0;
          Element(A,q,p) = 0;

          for( j = 0; j < n; ++j ) {
            cComplex x = Element(U,p,j);
            cComplex y = Element(U,q,j);
            Element(U,p,j) = x + sx*(y - tx*x);
            Element(U,q,j) = y - sy*(x + ty*y);
          }
        }
      }

    for( p = 0; p < n; ++p ) {
      ev[p][0] = 0;
      d[p] = ev[p][1];
    }
  }

  fputs("Bad convergence in CEigensystem\n", stderr);

done:
  if( sort == 0 ) return;

/* sort the eigenvalues by their real part */

  for( p = 0; p < n - 1; ++p ) {
    Integer j = p;
    Complex t = d[p];
    for( q = p + 1; q < n; ++q )
      if( sort*(creal(t) - creal(d[q])) > 0 ) t = d[j = q];
    if( j == p ) continue;
    d[j] = d[p];
    d[p] = t;
    for( q = 0; q < n; ++q ) {
      cComplex x = Element(U,p,q);
      Element(U,p,q) = Element(U,j,q);
      Element(U,j,q) = x;
    }
  }
}

