:Evaluate: BeginPackage["Diag`"]

:Evaluate: HEigensystem::usage =
	"HEigensystem[A] computes the eigenvalue decomposition {d, U} of the Hermitian matrix A fulfilling U.A == DiagonalMatrix[d].U with U unitary.
	HEigensystem[A, s] further determines the sorting: s = 1 ascending (default), -1 descending, 0 no sort."

:Evaluate: SEigensystem::usage =
	"SEigensystem[A] computes the eigenvalue decomposition {d, U} of the symmetric matrix A fulfilling U.A == DiagonalMatrix[d].U with Inverse[U] == Transpose[U].
	SEigensystem[A, s] further determines the sorting: s = 1 ascending (default), -1 descending, 0 no sort."

:Evaluate: CEigensystem::usage =
	"CEigensystem[A] computes the eigenvalue decomposition {d, U} of the general complex matrix A fulfilling U.A == DiagonalMatrix[d].U.
	CEigensystem[A, s] further determines the sorting: s = 1 ascending (default), -1 descending, 0 no sort."

:Evaluate: TakagiFactor::usage =
	"TakagiFactor[A] computes the Takagi factorization {d, U} of the symmetric matrix A fulfilling Conjugate[U].A == DiagonalMatrix[d].U with U unitary.
	TakagiFactor[A, s] further determines the sorting: s = 1 ascending (default), -1 descending, 0 no sort."

:Evaluate: SVD::usage =
	"SVD[A] computes the Singular Value Decomposition {V, d, W} of the matrix A fulfilling Conjugate[V].A == DiagonalMatrix[d].W.
	SVD[A, s] further determines the sorting: s = 1 ascending (default), -1 descending, 0 no sort."

:Evaluate: $NSweeps::usage =
	"$NSweeps returns the number of sweeps needed to diagonalize the matrix in the last invocation of HEigensystem, SEigensystem, CEigensystem, TakagiFactor, or SVD."

:Evaluate: Begin["`Private`"]

:Evaluate: matQ[m_] := MatrixQ[m, NumericQ]

:Evaluate: sqmatQ[m_] := matQ[m] && SameQ@@ Dimensions[m]

:Evaluate: pack[m_] := Flatten @ Map[{Re[#], Im[#]}&, N[m], {2}]

:Evaluate: cplx[f_] := Chop @ Apply[Complex, Partition[f, 2], 1]

:Evaluate: unpack[f_, n_] := Partition[cplx[f], n]

:Begin:
:Function: mHEigensystem
:Pattern: HEigensystem[A_List?sqmatQ, sort_Integer:1]
:Arguments: {Length[A], pack[A], sort}
:ArgumentTypes: {Integer, RealList, Integer}
:ReturnType: Manual
:End:

:Begin:
:Function: mSEigensystem
:Pattern: SEigensystem[A_List?sqmatQ, sort_Integer:1]
:Arguments: {Length[A], pack[A], sort}
:ArgumentTypes: {Integer, RealList, Integer}
:ReturnType: Manual
:End:

:Begin:
:Function: mCEigensystem
:Pattern: CEigensystem[A_List?sqmatQ, sort_Integer:1]
:Arguments: {Length[A], pack[A], sort}
:ArgumentTypes: {Integer, RealList, Integer}
:ReturnType: Manual
:End:

:Begin:
:Function: mTakagiFactor
:Pattern: TakagiFactor[A_List?sqmatQ, sort_Integer:1]
:Arguments: {Length[A], pack[A], sort}
:ArgumentTypes: {Integer, RealList, Integer}
:ReturnType: Manual
:End:

:Begin:
:Function: mSVD
:Pattern: SVD[A_List?matQ, sort_Integer:1]
:Arguments: {Length[A], Length[A[[1]]], pack[A], sort}
:ArgumentTypes: {Integer, Integer, RealList, Integer}
:ReturnType: Manual
:End:

:Begin:
:Function: mnsweeps
:Pattern: $NSweeps
:Arguments: {}
:ArgumentTypes: {}
:ReturnType: Integer
:End:

:Evaluate: End[]

:Evaluate: EndPackage[]



/*
	Diag.tm
		the Mathematica frontend for the libdiag routines
		last modified 19 Mar 10 th
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

#include "mathlink.h"
#ifndef MLCONST
#define MLCONST
#endif

#include "CDiag.h"

typedef MLCONST char cchar;
typedef const int cint;
typedef const long clong;
typedef double Real;
typedef const Real cReal;
typedef double_complex Complex;
typedef const Complex cComplex;

static inline int MLPutComplexList(MLINK mlp, cComplex *x, cint n)
{
  MLPutFunction(mlp, "Diag`Private`cplx", 1);
  return MLPutRealList(mlp, (Real *)x, 2*n);
}

static inline int MLPutComplexMatrix(MLINK mlp, cComplex *x, cint m, cint n)
{
  MLPutFunction(mlp, "Diag`Private`unpack", 2);
  MLPutRealList(mlp, (Real *)x, 2*m*n);
  return MLPutInteger(mlp, n);
}

/******************************************************************/

static void mHEigensystem(int n, Real *Alist, long Alen, int sort)
{
  Complex A[n*n], U[n*n];
  Real d[n];

  memcpy(A, Alist, Alen*sizeof(Real));

  HEigensystem(n, A,n, d, U,n, sort);

  MLPutFunction(stdlink, "List", 2);
  MLPutRealList(stdlink, d, n);
  MLPutComplexMatrix(stdlink, U, n,n);

  MLEndPacket(stdlink);
}

/******************************************************************/

static void mSEigensystem(int n, Real *Alist, long Alen, int sort)
{
  Complex A[n*n], U[n*n], d[n];

  memcpy(A, Alist, Alen*sizeof(Real));

  SEigensystem(n, A,n, d, U,n, sort);

  MLPutFunction(stdlink, "List", 2);
  MLPutComplexList(stdlink, d, n);
  MLPutComplexMatrix(stdlink, U, n,n);

  MLEndPacket(stdlink);
}

/******************************************************************/

static void mCEigensystem(int n, Real *Alist, long Alen, int sort)
{
  Complex A[n*n], U[n*n], d[n];

  memcpy(A, Alist, Alen*sizeof(Real));

  CEigensystem(n, A,n, d, U,n, sort);

  MLPutFunction(stdlink, "List", 2);
  MLPutComplexList(stdlink, d, n);
  MLPutComplexMatrix(stdlink, U, n,n);

  MLEndPacket(stdlink);
}

/******************************************************************/

static void mTakagiFactor(int n, Real *Alist, long Alen, int sort)
{
  Complex A[n*n], U[n*n];
  Real d[n];

  memcpy(A, Alist, Alen*sizeof(Real));

  TakagiFactor(n, A,n, d, U,n, sort);

  MLPutFunction(stdlink, "List", 2);
  MLPutRealList(stdlink, d, n);
  MLPutComplexMatrix(stdlink, U, n,n);

  MLEndPacket(stdlink);
}

/******************************************************************/

static void mSVD(int m, int n, Real *Alist, long Alen, int sort)
{
  cint nm = (m < n) ? m : n;
  Complex V[nm*m], W[nm*n];
  Real d[nm];

  SVD(m, n, (Complex *)Alist,n, d, V,m, W,n, sort);

  MLPutFunction(stdlink, "List", 3);
  MLPutComplexMatrix(stdlink, V, nm,m);
  MLPutRealList(stdlink, d, nm);
  MLPutComplexMatrix(stdlink, W, nm,n);

  MLEndPacket(stdlink);
}

/******************************************************************/

static int mnsweeps(void)
{
  return nsweeps;
}

/******************************************************************/

int main(int argc, char **argv)
{
  return MLMain(argc, argv);
}

