/*
   HEigensystem.c
   diagonalization of a Hermitian n-by-n matrix using the Jacobi algorithm
   code adapted from the "Handbook" routines for complex A
   (Wilkinson, Reinsch: Handbook for Automatic Computation, p. 202)
   this file is part of the Diag library
   last modified 19 Mar 10 th
*/

#include "diag-c.h"


/*
   HEigensystem diagonalizes a Hermitian n-by-n matrix.
   Input: n, A = n-by-n matrix, Hermitian
   (only the upper triangle of A needs to be filled).
   Output: d = vector of eigenvalues, U = transformation matrix
   these fulfill diag(d) = U A U^+ = U A U^-1 with U unitary.
*/

void HEigensystem(cInteger n, Complex *A, cInteger ldA,
  Real *d, Complex *U, cInteger ldU, cInteger sort)
{
  Integer p, q;
  cReal red = .04/(n*n*n*n);
  Real ev[n][2];

  for( p = 0; p < n; ++p ) {
    ev[p][0] = 0;
    ev[p][1] = d[p] = creal(Element(A,p,p));
  }

  for( p = 0; p < n; ++p ) {
    memset(&Element(U,p,0), 0, n*sizeof(Complex));
    Element(U,p,p) = 1;
  }

  for( nsweeps = 1; nsweeps <= 50; ++nsweeps ) {
    Real thresh = 0;
    for( q = 1; q < n; ++q )
      for( p = 0; p < q; ++p )
        thresh += csq(Element(A,p,q));
    if( !(thresh > SYM_EPS) ) goto done;

    thresh = (nsweeps < 4) ? thresh*red : 0;

    for( q = 1; q < n; ++q )
      for( p = 0; p < q; ++p ) {
        cComplex Apq = Element(A,p,q);
        cReal off = csq(Apq);
        cReal sqp = rsq(ev[p][1]);
        cReal sqq = rsq(ev[q][1]);
        if( nsweeps > 4 && off < SYM_EPS*max(sqp, sqq) )
          Element(A,p,q) = 0;
        else if( off > thresh ) {
          Real delta, t, s, invc;
          Integer j;

          t = .5*(ev[p][1] - ev[q][1]);
          t = 1/(t + copysign(sqrt(t*t + off), t));
          delta = off*t;
          ev[p][1] = d[p] + (ev[p][0] += delta);
          ev[q][1] = d[q] + (ev[q][0] -= delta);
          invc = sqrt(delta*t + 1);
          s = t/invc;
          t = delta/(invc + 1);

          for( j = 0; j < p; ++j ) {
            cComplex x = Element(A,j,p);
            cComplex y = Element(A,j,q);
            Element(A,j,p) = x + s*(conj(Apq)*y - t*x);
            Element(A,j,q) = y - s*(Apq*x + t*y);
          }

          for( j = p + 1; j < q; ++j ) {
            cComplex x = Element(A,p,j);
            cComplex y = Element(A,j,q);
            Element(A,p,j) = x + s*(Apq*conj(y) - t*x);
            Element(A,j,q) = y - s*(Apq*conj(x) + t*y);
          }

          for( j = q + 1; j < n; ++j ) {
            cComplex x = Element(A,p,j);
            cComplex y = Element(A,q,j);
            Element(A,p,j) = x + s*(Apq*y - t*x);
            Element(A,q,j) = y - s*(conj(Apq)*x + t*y);
          }

          Element(A,p,q) = 0;

          for( j = 0; j < n; ++j ) {
            cComplex x = Element(U,p,j);
            cComplex y = Element(U,q,j);
            Element(U,p,j) = x + s*(Apq*y - t*x);
            Element(U,q,j) = y - s*(conj(Apq)*x + t*y);
          }
        }
      }

    for( p = 0; p < n; ++p ) {
      ev[p][0] = 0;
      d[p] = ev[p][1];
    }
  }

  fputs("Bad convergence in HEigensystem\n", stderr);

done:
  if( sort == 0 ) return;

/* sort the eigenvalues */

  for( p = 0; p < n - 1; ++p ) {
    Integer j = p;
    Real t = d[p];
    for( q = p + 1; q < n; ++q )
      if( sort*(t - d[q]) > 0 ) t = d[j = q];
    if( j == p ) continue;
    d[j] = d[p];
    d[p] = t;
    for( q = 0; q < n; ++q ) {
      cComplex x = Element(U,p,q);
      Element(U,p,q) = Element(U,j,q);
      Element(U,j,q) = x;
    }
  }
}

