/*
   SVD.F
   singular value decomposition of an m-by-n matrix
   this file is part of the Diag library
   last modified 22 Mar 10 th
*/

#include "diag-c.h"


/*
   SVD performs a singular value decomposition.
   Input: m, n, A = m-by-n matrix.
   Output: d = nm-vector of singular values,
   V = nm-by-m left transformation matrix,
   W = nm-by-n right transformation matrix, nm = min(m, n),
   these fulfill diag(d) = V^* A W^+.
*/

void SVD(cInteger m, cInteger n, Complex *Ao, cInteger ldAo,
  Real *d, Complex *Vo, cInteger ldVo, Complex *Wo, cInteger ldWo,
  cInteger sort)
{
  Integer p, q, rev = ((m - n) >> 15) & 1;
  cInteger nm = imin(m, n);
  Integer pi[nm];
  cInteger nx = imax(m, n);
  cReal red = .01/(nx*nx*nx*nx);
  Complex VW[3][nx][nx];
  Complex *V = (Complex *)VW[0];
  Complex *W = (Complex *)VW[1];
  Complex *A = (Complex *)VW[2];
#define ldV nx
#define ldW nx
#define ldA nx

  memset(VW, 0, sizeof VW);
  for( p = 0; p < nx; ++p ) Element(V,p,p) = 1;
  for( p = 0; p < nx; ++p ) Element(W,p,p) = 1;

  if( rev )
    for( q = 0; q < m; ++q )
      for( p = 0; p < n; ++p )
        Element(A,p,q) = Element(Ao,q,p);
  else
    for( q = 0; q < m; ++q )
      for( p = 0; p < n; ++p )
        Element(A,q,p) = Element(Ao,q,p);

  for( nsweeps = 1; nsweeps <= 50; ++nsweeps ) {
    Real thresh = 0;
    for( q = 1; q < nx; ++q )
      for( p = 0; p < q; ++p )
        thresh += csq(Element(A,p,q)) + csq(Element(A,q,p));
    if( !(thresh > EPS) ) goto done;

    thresh = (nsweeps < 4) ? thresh*red : 0;

    for( q = 1; q < nx; ++q )
      for( p = 0; p < q; ++p ) {
        cComplex Apq = Element(A,p,q);
        cComplex Aqp = Element(A,q,p);
        cReal off = csq(Apq) + csq(Aqp);
        cComplex App = Element(A,p,p);
        cReal sqp = csq(App);
        cComplex Aqq = Element(A,q,q);
        cReal sqq = csq(Aqq);
        if( nsweeps > 4 && off < EPS*max(sqp, sqq) ) {
          Element(A,p,q) = 0;
          Element(A,q,p) = 0;
        }
        else if( off > thresh ) {
          Real xv, xw, dv, dw, t, invc;
          Complex sv, sw, tv, tw, evp, evq;
          Integer j;

          xv = creal((App - Aqq)*conj(App + Aqq));
          xw = creal((Apq - Aqp)*conj(Apq + Aqp));
          dv = .5*(xv + xw);
          dw = .5*(xv - xw);

          tv = conj(App)*Aqp + Aqq*conj(Apq);
          xv = sqrt(dv*dv + csq(tv));
          tw = conj(App)*Apq + Aqq*conj(Aqp);
          xw = sqrt(dw*dw + csq(tw));

          t = copysign(1, min(fabs(dv + xv), fabs(dw + xw)) -
                          min(fabs(dv - xv), fabs(dw - xw)));

          tv /= dv + t*xv;
          invc = sqrt(1 + csq(tv));
          sv = tv/invc;
          tv /= invc + 1;

          tw /= dw + t*xw;
          invc = sqrt(1 + csq(tw));
          sw = tw/invc;
          tw /= invc + 1;

          evp = invc*(App + conj(sv)*(Aqp - tv*App));
          evq = invc*(Aqq - sv*(Apq + conj(tv)*Aqq));

          for( j = 0; j < nx; ++j ) {
            Complex x = Element(A,j,p);
            Complex y = Element(A,j,q);
            Element(A,j,p) = x + conj(sw)*(y - tw*x);
            Element(A,j,q) = y - sw*(x + conj(tw)*y);
            x = Element(A,p,j);
            y = Element(A,q,j);
            Element(A,p,j) = x + conj(sv)*(y - tv*x);
            Element(A,q,j) = y - sv*(x + conj(tv)*y);
          }

          Element(A,p,p) = evp;
          Element(A,p,q) = 0;
          Element(A,q,p) = 0;
          Element(A,q,q) = evq;

          for( j = 0; j < nx; ++j ) {
            cComplex x = Element(V,p,j);
            cComplex y = Element(V,q,j);
            Element(V,p,j) = x + sv*(y - conj(tv)*x);
            Element(V,q,j) = y - conj(sv)*(x + tv*y);
          }

          for( j = 0; j < nx; ++j ) {
            cComplex x = Element(W,p,j);
            cComplex y = Element(W,q,j);
            Element(W,p,j) = x + sw*(y - conj(tw)*x);
            Element(W,q,j) = y - conj(sw)*(x + tw*y);
          }
        }
      }
  }

  fputs("Bad convergence in SVD\n", stderr);
          
done:

/* make the diagonal elements nonnegative */

  for( p = 0; p < nm; ++p ) {
    cComplex App = Element(A,p,p);
    d[p] = cabs(App);
    if( d[p] > DBL_EPS && d[p] != creal(App) ) {
      cComplex f = App/d[p];
      for( q = 0; q < nm; ++q ) Element(W,p,q) *= f;
    }
  }

/* sort the eigenvalues */

  for( p = 0; p < nm; ++p ) pi[p] = p;

  for( p = 0; p < nm; ++p ) {
    Integer j = p;
    Real t = d[p];
    if( sort )
      for( q = p + 1; q < n; ++q )
        if( sort*(t - d[q]) > 0 ) t = d[j = q];

    d[j] = d[p];
    d[p] = t;

    q = pi[j];
    pi[j] = pi[p];

    memcpy(&Element(Vo,p,0), VW[rev][q], m*sizeof(Complex));
    memcpy(&Element(Wo,p,0), VW[1-rev][q], n*sizeof(Complex));
  }
}

