/*
   TakagiFactor.F
   computes the Takagi factorization of a complex symmetric matrix
   code adapted from the "Handbook" routines
   (Wilkinson, Reinsch: Handbook for Automatic Computation, p. 202)
   this file is part of the Diag library
   last modified 19 Mar 10 th
*/

#include "diag-c.h"


/*
   TakagiFactor factorizes a complex symmetric n-by-n matrix
   Input: n, A = n-by-n matrix, complex symmetric
   (only the upper triangle of A needs to be filled).
   Output: d = vector of diagonal values, U = transformation matrix
   these fulfill diag(d) = U^* A U^+ with U unitary.
*/

void TakagiFactor(cInteger n, Complex *A, cInteger ldA,
  Real *d, Complex *U, cInteger ldU, cInteger sort)
{
  Integer p, q;
  cReal red = .04/(n*n*n*n);
  Complex ev[n][2];

  for( p = 0; p < n; ++p ) {
    ev[p][0] = 0;
    ev[p][1] = Element(A,p,p);
  }

  for( p = 0; p < n; ++p ) {
    memset(&Element(U,p,0), 0, n*sizeof(Complex));
    Element(U,p,p) = 1;
  }

  for( nsweeps = 1; nsweeps <= 50; ++nsweeps ) {
    Real thresh = 0;
    for( q = 1; q < n; ++q )
      for( p = 0; p < q; ++p )
        thresh += csq(Element(A,p,q));
    if( !(thresh > SYM_EPS) ) goto done;

    thresh = (nsweeps < 4) ? thresh*red : 0;

    for( q = 1; q < n; ++q )
      for( p = 0; p < q; ++p ) {
        cComplex Apq = Element(A,p,q);
        cReal off = csq(Apq);
        cReal sqp = csq(ev[p][1]);
        cReal sqq = csq(ev[q][1]);
        if( nsweeps > 4 && off < SYM_EPS*max(sqp, sqq) )
          Element(A,p,q) = 0;
        else if( off > thresh ) {
          Real t, invc;
          Complex f;
          Integer j;

          t = .5*fabs(sqp - sqq);
          if( t > DBL_EPS )
            f = copysign(1, sqp - sqq)*
              (ev[q][1]*conj(Apq) + conj(ev[p][1])*Apq);
          else
            f = (sqp == 0) ? 1 : csqrt(ev[q][1]/ev[p][1]);
          t += sqrt(t*t + csq(f));
          f /= t;

          ev[p][1] = Element(A,p,p) + (ev[p][0] += Apq*conj(f));
          ev[q][1] = Element(A,q,q) + (ev[q][0] -= Apq*f);

          t = csq(f);
          invc = sqrt(t + 1);
          f /= invc;
          t /= invc*(invc + 1);

          for( j = 0; j < p; ++j ) {
            cComplex x = Element(A,j,p);
            cComplex y = Element(A,j,q);
            Element(A,j,p) = x + (conj(f)*y - t*x);
            Element(A,j,q) = y - (f*x + t*y);
          }

          for( j = p + 1; j < q; ++j ) {
            cComplex x = Element(A,p,j);
            cComplex y = Element(A,j,q);
            Element(A,p,j) = x + (conj(f)*y - t*x);
            Element(A,j,q) = y - (f*x + t*y);
          }

          for( j = q + 1; j < n; ++j ) {
            cComplex x = Element(A,p,j);
            cComplex y = Element(A,q,j);
            Element(A,p,j) = x + (conj(f)*y - t*x);
            Element(A,q,j) = y - (f*x + t*y);
          }

          Element(A,p,q) = 0;

          for( j = 0; j < n; ++j ) {
            cComplex x = Element(U,p,j);
            cComplex y = Element(U,q,j);
            Element(U,p,j) = x + (f*y - t*x);
            Element(U,q,j) = y - (conj(f)*x + t*y);
          }
        }
      }

    for( p = 0; p < n; ++p ) {
      ev[p][0] = 0;
      Element(A,p,p) = ev[p][1];
    }
  }

  fputs("Bad convergence in TakagiFactor\n", stderr);

done:

/* make the diagonal elements nonnegative */

  for( p = 0; p < n; ++p ) {
    cComplex App = Element(A,p,p);
    d[p] = cabs(App);
    if( d[p] > DBL_EPS && d[p] != creal(App) ) {
      cComplex f = csqrt(App/d[p]);
      for( q = 0; q < n; ++q ) Element(U,p,q) *= f;
    }
  }

  if( sort == 0 ) return;

/* sort the eigenvalues */

  for( p = 0; p < n - 1; ++p ) {
    Integer j = p;
    Real t = d[p];
    for( q = p + 1; q < n; ++q )
      if( sort*(t - d[q]) > 0 ) t = d[j = q];
    if( j == p ) continue;
    d[j] = d[p];
    d[p] = t;
    for( q = 0; q < n; ++q ) {
      cComplex x = Element(U,p,q);
      Element(U,p,q) = Element(U,j,q);
      Element(U,j,q) = x;
    }
  }
}

